# -*- coding: utf-8 -*-
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution.

import sys
import os
import re
import time
import base64
import mimetypes
import urllib2

import db_backend

from EaseXML import *

KNOWN_USERS = {
    'testuser': {
        'name': 'Example User',
        'email': 'test@example.com',
        },
    }

RST_RE = r"""\{\{\{
#!rst
(.*)\}\}\}"""

PRE_RE = r"""\{\{\{
(.*)\}\}\}"""


class LP(Namespace):
    uri = 'https://launchpad.net/xmlns/2006/bugs'

    def getName(self):
        return ""


class Person(XMLObject):
    _name = 'person'
    username = StringAttribute(name='name')
    email = StringAttribute()
    text = TextNode(main=True)

    def set(self, nodeName, val):
        if nodeName == '_XO_name':
            val = val.lower()
            idx = val.find('@')
            if idx > 0:
                self.email = val
                val = val[:idx-1]
        XMLObject.set(self, nodeName, val)


class Reporter(Person):
    _name = 'reporter'


class Assignee(Person):
    _name = 'assignee'


class Sender(Person):
    _name = 'sender'


class Subscriber(Person):
    _name = 'subscriber'


class Attachment(XMLObject):
    _name = 'attachment'
    type = TextNode()
    filename = TextNode()
    title = TextNode()
    mimetype = TextNode()
    contents = TextNode()


class Comment(XMLObject):
    _name = 'comment'
    sender = ItemNode('Sender')
    date = TextNode()
    title = TextNode(optional=True)
    text = TextNode()
    attachments = ListNode('Attachment')


class Subscriptions(XMLObject):
    _name = 'subscriptions'
    persons = ListNode('Subscriber')


class Bug(XMLObject):
    _name = 'bug'
    id = StringAttribute()
    private = TextNode(optional=True)
    security_related = TextNode(optional=True)
    duplicateof = TextNode(optional=True)
    datecreated = TextNode()
    nickname = TextNode()
    title = TextNode()
    description = TextNode()
    reporter = ItemNode('Reporter')
    status = TextNode()
    importance = TextNode()
    milestone = TextNode(optional=True)
    assignee = ItemNode('Assignee', optional=True)
    comments = ListNode('Comment')
    subscriptions = ItemNode('Subscriptions')


class LaunchpadBugs(XMLObject):
    _name = 'launchpad-bugs'
    _prettyPrint = False
    bugs = ListNode('Bug')


def format_timestamp(timestamp):
    date = time.gmtime(timestamp)
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", date)


class TracDb:

    def __init__(self, filename, attachments_dir):
        self._db = db_backend.DBBackend(db_backend='sqlite3',
                                        database=filename)
        self._db.connect()
        self._users = {}
        self._attachments_dir = attachments_dir

        self._known_users_by_email = {}
        for username, data in KNOWN_USERS.iteritems():
            email = data['email']
            name = data['name']
            new_data = {'name': name, 'username': username,
                        'email': email}
            self._known_users_by_email[email] = new_data

    def get_user(self, username):
        user = {}
        is_email = username.find('@') > -1
        if is_email:
            email = username

            try:
                user = self._known_users_by_email[email]
            except KeyError:
                query = """
                    SELECT sid AS username FROM session_attribute
                    WHERE name='email' AND value=?
                """
                rows = self._db.sql_execute(query, email)
                if len(rows) > 0:
                    username = rows[0].username

        if not user:
            try:
                user = KNOWN_USERS[username]
            except KeyError:
                try:
                    user = self._users[username]
                except KeyError:
                    user = {'username': username, 'email': '',
                            'name': ''}

                    rows = self._db.sql_execute(
                        "SELECT * FROM session_attribute "
                        "WHERE sid = ?", username)
                    for row in rows:
                        if row.name == 'email':
                            user['email'] = row.value
                        elif row.name == 'name':
                            user['name'] = row.value

                    self._users[username] = user

        return user

    def get_tickets(self, count=None):
        query = "select * from ticket where reporter!=''"
        if count is not None:
            query += " limit %s" % count
        rows = self._db.sql_execute(query)
        return rows

    def get_comments(self, ticket):
        ticket_id = ticket['id']
        query = "select time,author,newvalue text "\
                "from ticket_change where ticket=? "\
                "and field='comment' and newvalue != '' "\
                "order by time"
        rows = self._db.sql_execute(query, ticket_id)
        return rows

    def get_attachments(self, ticket):
        ticket_id = ticket['id']
        query = (
            "SELECT time, author, filename, description AS text "
            "FROM attachment WHERE id = ? "
            "AND type = 'ticket' "
            "ORDER BY time")
        rows = [
            d.as_dict() for d in self._db.sql_execute(query, ticket_id)]

        for row in rows:
            # TODO: read file
            filename = urllib2.quote(row['filename'].encode('utf8'))

            attachment_filename = os.path.join(
                self._attachments_dir, str(ticket_id), filename)

            data = open(attachment_filename).read()
            row['data']  = data

        return rows


def convert_status_importance(ticket):
    trac_status = ticket['status']
    trac_priority = ticket['priority']
    #trac_severity = ticket['severity']
    trac_resolution = ticket['resolution']

    priority_to_importance = {
        'high': 'HIGH',
        'highest': 'CRITICAL',
        'low': 'LOW',
        'normal': 'MEDIUM',
        'unassigned': 'UNDECIDED',
        'wishlist': 'WISHLIST',
        }
    importance = priority_to_importance.get(
        trac_priority, 'UNKNOWN')

    if trac_status == 'closed':
        resolution_to_status = {
            '': 'NEW',
            'fixed': 'FIXRELEASED',
            'invalid': 'INVALID',
            'wontfix': 'WONTFIX',
            }
        status = resolution_to_status.get(
            trac_resolution, 'UNKNOWN')
    else:
        status_dict = {
            'new': 'NEW', 'reopened': 'NEW',
            'assigned': 'INPROGRESS'}
        status = status_dict.get(trac_status, 'UNKNOWN')
    return status, importance


def _rstrip(line, JUNK='\n \t'):
    """Return line stripped of trailing spaces, tabs, newlines.

    Note that line.rstrip() instead also strips sundry control
    characters, but at least one known Emacs user expects to keep junk
    like that, not mentioning Barry by name or anything <wink>.
    """

    i = len(line)
    while i > 0 and line[i-1] in JUNK:
        i -= 1
    return line[:i]


def render_text(text):
    # handle trac {{{ }}} formatting (incl. ReST).
    text = ''.join([_rstrip(line).expandtabs() + "\n"
                    for line in text.splitlines()])
    for regex in (RST_RE, PRE_RE):
        rg = re.compile(regex,
                        re.IGNORECASE|re.DOTALL|re.MULTILINE)
        m = rg.search(text)
        if m:
            text = m.group(1)
            break

    return text


def create_bug(db, trac_ticket):
    bug_id = trac_ticket['id']
    bug_type = trac_ticket['type']
    datecreated = format_timestamp(int(trac_ticket['time']))
    title = trac_ticket['summary']
    description = render_text(trac_ticket['description'])
    reporter = trac_ticket['reporter']
    milestone = trac_ticket['milestone']
    owners = trac_ticket['owner']
    subscribers = trac_ticket['cc']

    trac_reporter = db.get_user(reporter)
    #reporter_email = trac_reporter.get('email', 'foo@bar.com')
    reporter_email = trac_reporter['email']
    if not reporter_email:
        print '[ticket %s] pb with reporter %s' % (bug_id, reporter)
        return None

    reporter_name = trac_reporter.get('name', 'John Doe')
    lp_reporter = Reporter(username=reporter,
                           email=reporter_email,
                           text=reporter_name)

    status, importance = convert_status_importance(trac_ticket)

    assignee = None
    if owners is not None:
        owner = owners.split(',')[0]
        if owner:
            trac_owner = db.get_user(owner)
            owner_email = trac_owner['email']
            if not owner_email:
                print '[ticket %s] pb with owner %s' % (bug_id,
                                                        owner)
                return None

            owner_name = trac_owner.get('name', 'John Doe')

            assignee = Assignee(username=owner, email=owner_email,
                                text=owner_name)

    lp_subscribers = []
    if subscribers:
        for username in subscribers.split(','):
            username = username.strip()
            trac_person = db.get_user(username)
            person_email = trac_person['email']
            if not person_email:
                print '[ticket %s] pb with cc %s' % (bug_id,
                                                     username)
                continue
            lp_subscribers.append(Subscriber(username=username,
                                             email=person_email,
                                             text=trac_person['name']))
    subscriptions = Subscriptions(persons=lp_subscribers)

    contents = dict(id=str(bug_id),
                    datecreated=datecreated,
                    nickname="elisa-%s" % bug_id,
                    title=title,
                    description=description,
                    reporter=lp_reporter,
                    status=status,
                    importance=importance,
                    subscriptions=subscriptions)
    if assignee:
        contents['assignee'] = assignee

    if milestone:
        invalid = {'0.0.1 Guadec06 Preview': '0.0.1',
                   '0.0 bugfix': '0.0.1',
                   'elisa-0.3.3': 'core-0.3.3',
                   'testing': '',
                   'elisa-0.3.6': 'core-0.3.6',
                   'elisa-0.3.4': 'core-0.3.4',
                   'misc.win32.installer-0.1': '',
                   'elisa-0.3.5': 'core-0.3.5',
                   }
        try:
            milestone = invalid[milestone]
        except KeyError:
            pass
        else:
            if milestone:
                milestone = "elisa-%s" % milestone
                contents['milestone'] = milestone

    bug = Bug(**contents)

    sender = Sender(username=reporter, email=reporter_email,
                    text=reporter_name)
    first_comment = Comment(sender=sender, date=datecreated,
                            title=title, text=description)
    #print first_comment.toXml()
    bug.comments.append(first_comment)
    return bug

def create_comment(db, trac_comment):
    datecreated = format_timestamp(int(trac_comment['time']))

    description = render_text(trac_comment['text'])
    #title = "%s ..." % description[:20]

    reporter = trac_comment['author']
    trac_reporter = db.get_user(reporter)
    reporter_email = trac_reporter['email']
    if not reporter_email:
        print '[comment %s] pb with comment reporter %s' % (1,
                                                            reporter)
        return None
    reporter_name = trac_reporter.get('name', 'John Doe')
    sender = Sender(username=reporter, email=reporter_email,
                    text=reporter_name)
    lp_comment = Comment(sender=sender, date=datecreated,
                         #title=title,
                         text=description)

    # is there an attachment?
    if trac_comment.has_key('filename'):
        filename = trac_comment['filename']
        mime, typ = mimetypes.guess_type(filename)
        if mime is None:
            mime = ""
        contents = base64.encodestring(trac_comment['data'])
        attachment = Attachment(type="UNSPECIFIED",
                                filename=filename,
                                title=trac_comment['text'],
                                mimetype=mime,
                                contents=contents)
        lp_comment.attachments.append(attachment)

    return lp_comment

def main(args):
    trac_db = args[0]
    attachments_dir = args[1]
    filename = args[2]

    try:
        count = args[3]
    except IndexError:
        count = None

    bugs = LaunchpadBugs()

    db = TracDb(trac_db, attachments_dir)
    for ticket in db.get_tickets(count):
        bug = create_bug(db, ticket)
        if not bug:
            continue

        # comments
        for comment in db.get_comments(ticket):
            lp_comment = create_comment(db, comment)
            if lp_comment:
                bug.comments.append(lp_comment)

        # attachments
        for attachment in db.get_attachments(ticket):
            lp_comment = create_comment(db, attachment)
            if lp_comment:
                bug.comments.append(lp_comment)

        bugs.bugs.append(bug)

    xml = bugs.toXml()

    fd = open(filename, 'w')
    fd.write(xml)
    fd.close()

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
